/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include "support/PortableMacros.h"
#include "syslog/SysLog.h"

#ifdef WATCH_MUSCLE_MEM_USAGE

#include "hashtable/PortableHashtable.h"
#include "string/PortableString.h"
#include "netutil/NetworkUtilityFunctions.h"

static PortableHashtable<PortableString, uint64> * _allocTable = NULL;

static struct timeval _nextPrintAt = {0, 0};
static bool _isNotingAllocation = false;  // avoids infinite regress when NoteAllocated uses hashtable, string, etc.
static uint32 _logID = 0;

void NoteAllocated(void * This, const char * className)
{
   if (_isNotingAllocation == false)
   {
      bool printLog = false;

      _isNotingAllocation = true;
      {
         const PortableString cs(className); 

         uint64 prevVal = 0LL;
         if (_allocTable == NULL) _allocTable = new PortableHashtable<PortableString, uint64>(PortableStringHashFunc);
         (void) _allocTable->Get(cs, prevVal);
         _allocTable->Put(cs, prevVal+1);

         LogTime(LOG_TRACE, "@Alloced obj=%p (%s) (%i)\n", This, className, prevVal+1);

         // Print current state to stdout every so many seconds
         struct timeval now; gettimeofday(&now, NULL);
         if (!IsLessThan(now, _nextPrintAt))
         {
            printLog = true;
            _nextPrintAt = now; 
   
            struct timeval interval = {WATCH_MUSCLE_MEM_USAGE, 0};
            AddTimeVal(_nextPrintAt, interval);
         }
      }
      _isNotingAllocation = false;
      if (printLog) LogCurrentAllocations();
   }
}

void NoteDeleted(void * This, const char * className)
{
   if (_isNotingAllocation == false)
   {
      _isNotingAllocation = true;
      {
         uint64 prevVal;
         PortableString cs(className);
         if ((_allocTable)&&(_allocTable->Get(cs, prevVal)&&(prevVal > 0)))
         {
            if (prevVal > 1) _allocTable->Put(cs, prevVal-1);
                        else _allocTable->Remove(cs);

            LogTime(LOG_TRACE, "@Deleted obj=%p (%s) (%i)\n", This, className, prevVal-1);
            if (_allocTable->Size() == 0)
            {
               delete _allocTable;
               _allocTable = NULL;
            }
         }
         else
         {
            PortableString errMsg("NoteDeleted:  Objects of class ");
            errMsg += className;
            errMsg += " deleted too many times!?";
            ASSERT(false, errMsg.Cstr());
         }
      }
      _isNotingAllocation = false;
   }
}

void LogCurrentAllocations()
{
   if (_isNotingAllocation == false)
   {
      _logID++;
      _isNotingAllocation = true;
      LogTime(LOG_INFO, "=====CURRENT OBJECT STATISTICS (%i)=======\n", _allocTable?_allocTable->Size():0);
      if (_allocTable)
      {
         PortableHashtableIterator<PortableString, uint64> iter = _allocTable->GetIterator();
         PortableString * nextKey;
         while((nextKey = iter.GetNextKey()) != NULL) LogTime(LOG_INFO, "  %s: %Lu\n", nextKey->Cstr(), *iter.GetNextValue());
      }
      _isNotingAllocation = false;
   }
}

#endif  // WATCH_MUSCLE_MEM_USAGE

static int _fileLogLevel    = LOG_NONE;
static int _consoleLogLevel = LOG_INFO;
static FILE * _logFile      = NULL;

static const char * _logLevelNames[] = {
   "None",
   "Critical Errors Only",
   "Errors Only",
   "Warnings and Errors Only",
   "Informational",
   "Debug",
   "Trace"
};

// This code put a #define due to the difficulty in making into a function
// (it's hard to forward varargs, ya know?)
#define LOG_BODY                              \
{                                             \
   if ((ll <= _fileLogLevel)&&(_logFile == NULL)) \
   {                                          \
      time_t n = time(NULL);                  \
      struct tm * now = localtime(&n);        \
      char tb[100];                           \
      sprintf(tb, "%02i%02i%02i%02i.log", now->tm_mon+1, now->tm_mday, now->tm_hour, now->tm_min); \
      _logFile = fopen(tb, "w");              \
      if (_logFile) LogTime(LOG_INFO, "Created log file %s\n", tb); \
   }                                          \
   if (ll <= _consoleLogLevel)                \
   {                                          \
      va_list argList;                        \
      va_start(argList, fmt);                 \
      vprintf(fmt, argList);                  \
   }                                          \
   if ((ll <= _fileLogLevel)&&(_logFile))     \
   {                                          \
      va_list argList;                        \
      va_start(argList, fmt);                 \
      vfprintf(_logFile, fmt, argList);       \
      fflush(_logFile);                       \
   }                                          \
}

const char * GetLogLevelName(int ll)
{
   return ((ll>=0)&&(ll<ARRAYITEMS(_logLevelNames))) ? _logLevelNames[ll] : "???";
}

int GetFileLogLevel()
{
   return _fileLogLevel;
}

int GetConsoleLogLevel()
{
   return _consoleLogLevel;
}

int GetMaxLogLevel()
{
   return (_consoleLogLevel > _fileLogLevel) ? _consoleLogLevel : _fileLogLevel;
}

void SetFileLogLevel(int loglevel)
{
   _fileLogLevel = loglevel;
   LogTime(LOG_INFO, "File logging level set to: %s\n", GetLogLevelName(_fileLogLevel));
}

void SetConsoleLogLevel(int loglevel)
{
   _consoleLogLevel = loglevel;
   LogTime(LOG_INFO, "Console logging level set to: %s\n", GetLogLevelName(_consoleLogLevel));
}

void LogTime(int ll, const char * fmt, ...)
{
   time_t n = time(NULL);
   struct tm * now = localtime(&n);
   char timebuf[100];
   sprintf(timebuf, "[%c %02i/%02i %02i:%02i:%02i] ", GetLogLevelName(ll)[0], now->tm_mon+1, now->tm_mday, now->tm_hour, now->tm_min, now->tm_sec);
   Log(ll, timebuf);
   LOG_BODY;
}

void Log(int ll, const char * fmt, ...)
{
   LOG_BODY;
}


